@@ -37,11 +37,10 @@ warnings.filterwarnings(
 
 import distro_info
 import fnmatch
+import logging
 import os
 import subprocess
 import sys
-import threading
-import time
 from gettext import gettext as _
 
 import dbus
@@ -72,6 +71,7 @@ from .Core.UpdateList import UpdateList
 from .Core.utils import get_arch, get_dist, SoftwarePropertiesPage
 from .backend import InstallBackend, get_backend
 
+IO_ERR_QUARK = Gio.io_error_quark()
 # file that signals if we need to reboot
 REBOOT_REQUIRED_FILE = "/var/run/reboot-required"
 
@@ -189,9 +189,6 @@ class UpdateManager(Gtk.Window):
         return self.close()
 
     def close(self):
-        if not self.get_sensitive():
-            return True
-
         if self.controller:
             controller_close = self.controller.close()
             if controller_close:
@@ -210,18 +207,14 @@ class UpdateManager(Gtk.Window):
         if "WAYLAND_DISPLAY" not in os.environ:
             cmd += ["--toplevel", "%s" % self.get_window().get_xid()]
 
-        self._look_busy()
         try:
-            p = subprocess.Popen(cmd)
-        except OSError:
-            pass
+            p = Gio.Subprocess.new(cmd, Gio.SubprocessFlags.NONE)
+        except Exception as e:
+            logging.exception(e)
         else:
-            while p.poll() is None:
-                while Gtk.events_pending():
-                    Gtk.main_iteration()
-                time.sleep(0.05)
-        finally:
-            self.start_available()
+            self._look_busy()
+            p.wait_check_async(None, lambda _a, _b, _c: self.start_available(),
+                               None)
 
     def start_update(self):
         if self.options.no_update:
@@ -242,19 +235,131 @@ class UpdateManager(Gtk.Window):
         self._start_pane(install_backend)
 
     def start_available(self, cancelled_update=False, error_occurred=False):
+        def refresh_cache(task, source_obj, data, cancellable):
+            try:
+                if self.cache is None:
+                    self.cache = MyCache(None)
+                else:
+                    self.cache.open(None)
+                    self.cache._initDepCache()
+            except AssertionError as e:
+                task.return_error(GLib.Error.new_literal(
+                    IO_ERR_QUARK, str(e), Gio.IOErrorEnum.FAILED
+                ))
+                return
+            except SystemError as e:
+                task.return_error(GLib.Error.new_literal(
+                    IO_ERR_QUARK, str(e), Gio.IOErrorEnum.INVALID_DATA
+                ))
+                return
+
+            self._check_oem_metapackages()
+
+            self._get_ua_security_status()
+
+            for pkgname in self.oem_metapackages:
+                try:
+                    if not self.cache[pkgname].is_installed:
+                        self.cache[pkgname].mark_install()
+                except SystemError:
+                    pass
+
+            self.update_list = UpdateList(self)
+            try:
+                self.update_list.update(
+                    self.cache,
+                    duplicate_packages=self.duplicate_packages,
+                    ua_security_packages=self.ua_security_packages,
+                )
+            except SystemError as e:
+                task.return_error(GLib.Error.new_literal(
+                    IO_ERR_QUARK, str(e), Gio.IOErrorEnum.NOT_SUPPORTED)
+                )
+                return
+
+            if self.update_list.distUpgradeWouldDelete > 0:
+                task.return_error(GLib.Error.new_literal(
+                    IO_ERR_QUARK, "", Gio.IOErrorEnum.PARTIAL_INPUT),
+                )
+                return
+            task.return_value()
+
         self._look_busy()
-        self.refresh_cache()
 
-        if self.cache is None:
+        t = Gio.Task.new(self, None, self.cache_refreshed,
+                         (cancelled_update, error_occurred))
+        t.run_in_thread(refresh_cache)
+
+    def cache_refreshed(self, source, task, data):
+        cancelled_update, error_occurred = data
+        try:
+            task_returned, task_result = task.propagate_value()
+        except Exception as e:
+            # FIXME: An unexpected thread exception will not be displayed
+            if not isinstance(e, GLib.Error):
+                header = _("Failed to refresh cache")
+                desc = _(
+                    "Refreshing the cache failed.\n\n"
+                    "Please report this bug against the 'update-manager' "
+                    "package and include the thread traceback from the "
+                    "terminal.\n"
+                )
+                self.start_error(False, header, desc)
+
+            if e.matches(IO_ERR_QUARK, Gio.IOErrorEnum.FAILED):
+                # if the cache could not be opened for some reason,
+                # let the release upgrader handle it, it deals
+                # a lot better with this
+                self._start_pane(PartialUpgradeDialog(self))
+                # we assert a clean cache
+                header = _("Software index is broken")
+                desc = _(
+                    "It is impossible to install or remove any software. "
+                    'Please use the package manager "Synaptic" or run '
+                    '"sudo apt-get install -f" in a terminal to fix '
+                    "this issue at first."
+                )
+                self.start_error(False, header, desc)
+            elif e.matches(IO_ERR_QUARK, Gio.IOErrorEnum.INVALID_DATA):
+                header = _("Could not initialize the package information")
+                desc = _(
+                    "An unresolvable problem occurred while "
+                    "initializing the package information.\n\n"
+                    "Please report this bug against the 'update-manager' "
+                    "package and include the following error message:\n"
+                ) + e.message
+                self.start_error(True, header, desc)
+            elif e.matches(IO_ERR_QUARK, Gio.IOErrorEnum.NOT_SUPPORTED):
+                header = _("Could not calculate the upgrade")
+                desc = _(
+                    "An unresolvable problem occurred while "
+                    "calculating the upgrade.\n\n"
+                    "Please report this bug against the 'update-manager' "
+                    "package and include the following error message:\n"
+                ) + e.message
+                self.start_error(True, header, desc)
+            elif e.matches(IO_ERR_QUARK, Gio.IOErrorEnum.PARTIAL_INPUT):
+                self._start_pane(PartialUpgradeDialog(self))
+            else:
+                header = _("Failed to refresh cache")
+                desc = _(
+                    "Refreshing the cache failed.\n\n"
+                    "Please report this bug against the 'update-manager' "
+                    "package and include the following error message:\n"
+                ) + e.message
+                self.start_error(False, header, desc)
             return
+        else:
+            if self.cache is None:
+                return
 
-        pane = self._make_available_pane(
-            self.cache.install_count + self.cache.del_count,
-            os.path.exists(REBOOT_REQUIRED_FILE),
-            cancelled_update,
-            error_occurred,
-        )
-        self._start_pane(pane)
+            pane = self._make_available_pane(
+                self.cache.install_count + self.cache.del_count,
+                os.path.exists(REBOOT_REQUIRED_FILE),
+                cancelled_update,
+                error_occurred,
+            )
+            self._start_pane(pane)
 
     def _check_oem_metapackages(self):
         di = distro_info.UbuntuDistroInfo()
@@ -277,22 +382,21 @@ class UpdateManager(Gtk.Window):
             if fnmatch.fnmatch(pkg.name, "oem-*-meta") and pkg.installed:
                 self.oem_metapackages.add(pkg)
 
-    def _fetch_ua_updates(self):
+    def _fetch_ua_updates(self, cb=None):
         try:
             self.ua_updates = ua.updates().updates
         except Exception as e:
             print("Error running updates end-point: ", e)
             self.ua_updates = []
+        if cb:
+            cb()
 
     def _get_ua_security_status(self):
         self.ua_security_packages = []
         self.ua_updates = []
-        t = threading.Thread(target=self._fetch_ua_updates, daemon=True)
-        t.start()
-        while t.is_alive():
-            while Gtk.events_pending():
-                Gtk.main_iteration()
-            time.sleep(0.05)
+        GLib.idle_add(self.sift_pro_packages)
+
+    def sift_pro_packages(self):
         for package in self.ua_updates:
             if (
                 package.provided_by == 'standard-security'
@@ -440,79 +544,6 @@ class UpdateManager(Gtk.Window):
                 # for debugging
                 # print(self.hwe_replacement_packages)
 
-    # fixme: we should probably abstract away all the stuff from libapt
-    def refresh_cache(self):
-        try:
-            if self.cache is None:
-                self.cache = MyCache(None)
-            else:
-                self.cache.open(None)
-                self.cache._initDepCache()
-        except AssertionError:
-            # if the cache could not be opened for some reason,
-            # let the release upgrader handle it, it deals
-            # a lot better with this
-            self._start_pane(PartialUpgradeDialog(self))
-            # we assert a clean cache
-            header = _("Software index is broken")
-            desc = _(
-                "It is impossible to install or remove any software. "
-                'Please use the package manager "Synaptic" or run '
-                '"sudo apt-get install -f" in a terminal to fix '
-                "this issue at first."
-            )
-            self.start_error(False, header, desc)
-            return
-        except SystemError as e:
-            header = _("Could not initialize the package information")
-            desc = _(
-                "An unresolvable problem occurred while "
-                "initializing the package information.\n\n"
-                "Please report this bug against the 'update-manager' "
-                "package and include the following error message:\n"
-            ) + str(e)
-            self.start_error(False, header, desc)
-            return
-
-        # Let the Gtk event loop breath if it hasn't had a chance.
-        def iterate():
-            while Gtk.events_pending():
-                Gtk.main_iteration()
-
-        iterate()
-
-        self._check_oem_metapackages()
-
-        self._get_ua_security_status()
-
-        for pkgname in self.oem_metapackages:
-            try:
-                if not self.cache[pkgname].is_installed:
-                    self.cache[pkgname].mark_install()
-            except SystemError:
-                pass
-
-        self.update_list = UpdateList(self)
-        try:
-            self.update_list.update(
-                self.cache,
-                eventloop_callback=iterate,
-                duplicate_packages=self.duplicate_packages,
-                ua_security_packages=self.ua_security_packages,
-            )
-        except SystemError as e:
-            header = _("Could not calculate the upgrade")
-            desc = _(
-                "An unresolvable problem occurred while "
-                "calculating the upgrade.\n\n"
-                "Please report this bug against the 'update-manager' "
-                "package and include the following error message:\n"
-            ) + str(e)
-            self.start_error(True, header, desc)
-
-        if self.update_list.distUpgradeWouldDelete > 0:
-            self._start_pane(PartialUpgradeDialog(self))
-
     def _setup_dbus(self):
         """this sets up a dbus listener if none is installed already"""
         # check if there is another g-a-i already and if not setup one
